@@ -23,6 +23,7 @@ def fabCommand(f):
     help="Comma separated list of component fields field with LCSC order code. First existing field is used")
 @click.option("--corrections", type=str, default="JLCPCB_CORRECTION",
     help="Comma separated list of component fields with the correction value. First existing field is used")
+@click.option("--correctionpatterns", type=click.Path(dir_okay=False))
 @click.option("--missingError/--missingWarn", help="If a non-ignored component misses LCSC field, fail")
 def jlcpcb(**kwargs):
     """
@@ -32,7 +33,13 @@ def jlcpcb(**kwargs):
     from kikit.common import fakeKiCADGui
     app = fakeKiCADGui()
 
-    return jlcpcb.exportJlcpcb(**kwargs)
+    try:
+        return jlcpcb.exportJlcpcb(**kwargs)
+    except Exception as e:
+        import sys
+        sys.stderr.write("An error occurred: " + str(e) + "\n")
+        sys.stderr.write("No output files produced\n")
+        sys.exit(1)
 
 
 @click.command()
@@ -42,6 +49,7 @@ def jlcpcb(**kwargs):
 @click.option("--ignore", type=str, default="", help="Comma separated list of designators to exclude from SMT assembly")
 @click.option("--corrections", type=str, default="PCBWAY_CORRECTION",
     help="Comma separated list of component fields with the correction value. First existing field is used")
+@click.option("--correctionpatterns", type=click.Path(dir_okay=False))
 @click.option("--manufacturer", type=str, default="Manufacturer",
     help="Comma separated list of fields to extract manufacturer name from. First existing field is used.")
 @click.option("--partNumber", type=str, default="PartNumber",
@@ -65,7 +73,13 @@ def pcbway(**kwargs):
     from kikit.common import fakeKiCADGui
     app = fakeKiCADGui()
 
-    return pcbway.exportPcbway(**kwargs)
+    try:
+        return pcbway.exportPcbway(**kwargs)
+    except Exception as e:
+        import sys
+        sys.stderr.write("An error occurred: " + str(e) + "\n")
+        sys.stderr.write("No output files produced\n")
+        sys.exit(1)
 
 
 @click.command()
@@ -78,7 +92,13 @@ def oshpark(**kwargs):
     from kikit.common import fakeKiCADGui
     app = fakeKiCADGui()
 
-    return oshpark.exportOSHPark(**kwargs)
+    try:
+        return oshpark.exportOSHPark(**kwargs)
+    except Exception as e:
+        import sys
+        sys.stderr.write("An error occurred: " + str(e) + "\n")
+        sys.stderr.write("No output files produced\n")
+        sys.exit(1)
 
 
 @click.group()
