@@ -63,6 +63,24 @@ class Section(click.ParamType):
                 param,
                 ctx)
 
+class HookPlugin(click.ParamType):
+    """
+    A CLI argument type for a HookPlugin. The format is <moduleName or
+    path>:<plugin name>:<arg>. The arg is optional.
+    """
+    name = "<module>.<plugin>:[arg]"
+
+    def convert(self, value, param, ctx):
+        pieces = value.split(":", maxsplit=1)
+        specPieces = pieces[0].rsplit(".", maxsplit=1)
+        if len(specPieces) < 2:
+            self.fail(f"{value} is not a valid plugin specification")
+        module = specPieces[0]
+        pluginName = specPieces[1]
+        arg = "" if len(pieces) == 2 else pieces[1]
+        return (module, pluginName, arg)
+
+
 def completePath(prefix, fileSuffix=""):
     """
     This is rather hacky and  far from ideal, however, until Click 8 we probably
@@ -127,6 +145,9 @@ def completeSection(section):
 @click.option("--preset", "-p", multiple=True,
     help="A panelization preset file; use prefix ':' for built-in styles.",
     **addCompatibleShellCompletion(completePreset))
+@click.option("--plugin", multiple=True, type=HookPlugin(),
+    help="A hook plugin to use during the panelization",
+    **addCompatibleShellCompletion(completePreset))
 @click.option("--layout", "-l", type=Section(),
     help="Override layout settings.",
     **addCompatibleShellCompletion(completeSection(LAYOUT_SECTION)))
@@ -151,6 +172,18 @@ def completeSection(section):
 @click.option("--text", "-t", type=Section(),
     help="Override text settings.",
     **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
+@click.option("--text2", type=Section(),
+    help="Override text settings.",
+    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
+@click.option("--text3", type=Section(),
+    help="Override text settings.",
+    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
+@click.option("--text4", type=Section(),
+    help="Override text settings.",
+    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
+@click.option("--copperfill", "-u", type=Section(),
+    help="Override copper fill settings.",
+    **addCompatibleShellCompletion(completeSection(COPPERFILL_SECTION)))
 @click.option("--page", "-P", type=Section(),
     help="Override page settings.",
     **addCompatibleShellCompletion(completeSection(POST_SECTION)))
@@ -162,8 +195,9 @@ def completeSection(section):
     **addCompatibleShellCompletion(completeSection(DEBUG_SECTION)))
 @click.option("--dump", "-d", type=click.Path(file_okay=True, dir_okay=False),
     help="Dump constructured preset into a JSON file.")
-def panelize(input, output, preset, layout, source, tabs, cuts, framing,
-                tooling, fiducials, text, page, post, debug, dump):
+def panelize(input, output, preset, plugin, layout, source, tabs, cuts, framing,
+             tooling, fiducials, text, text2, text3, text4, copperfill, page,
+             post, debug, dump):
     """
     Panelize boards
     """
@@ -176,10 +210,11 @@ def panelize(input, output, preset, layout, source, tabs, cuts, framing,
 
         preset = ki.obtainPreset(preset,
             layout=layout, source=source, tabs=tabs, cuts=cuts, framing=framing,
-            tooling=tooling, fiducials=fiducials, text=text, page=page, post=post,
-            debug=debug)
+            tooling=tooling, fiducials=fiducials, text=text,text2=text2,
+            text3=text3, text4=text4, copperfill=copperfill, page=page,
+            post=post, debug=debug)
 
-        doPanelization(input, output, preset)
+        doPanelization(input, output, preset, plugin)
 
         if (dump):
             with open(dump, "w") as f:
@@ -192,7 +227,7 @@ def panelize(input, output, preset, layout, source, tabs, cuts, framing,
             traceback.print_exc(file=sys.stderr)
         sys.exit(1)
 
-def doPanelization(input, output, preset):
+def doPanelization(input, output, preset, plugins=[]):
     """
     The panelization logic is separated into a separate function so we can
     handle errors based on the context; e.g., CLI vs GUI
@@ -205,43 +240,65 @@ def doPanelization(input, output, preset):
 
     if preset["debug"]["deterministic"] and isV6():
         pcbnew.KIID.SeedGenerator(42)
+    if preset["debug"]["drawtabfail"]:
+        import kikit.substrate
+        kikit.substrate.TABFAIL_VISUAL = True
 
     board = LoadBoard(input)
-
     panel = Panel(output)
+
+    useHookPlugins = ki.loadHookPlugins(plugins, board, preset)
+
+    useHookPlugins(lambda x: x.prePanelSetup(panel))
+
+    # Register extra footprints for annotations
+    for tabFootprint in preset["tabs"]["tabfootprints"]:
+        panel.annotationReader.registerTab(tabFootprint.lib, tabFootprint.footprint)
+
     panel.inheritDesignSettings(board)
     panel.inheritProperties(board)
     panel.inheritTitleBlock(board)
 
+    useHookPlugins(lambda x: x.afterPanelSetup(panel))
+
     sourceArea = ki.readSourceArea(preset["source"], board)
-    substrates = ki.buildLayout(preset["layout"], panel, input, sourceArea)
-    framingSubstrates = ki.dummyFramingSubstrate(substrates,
-        ki.frameOffset(preset["framing"]))
-    panel.buildPartitionLineFromBB(framingSubstrates)
+    substrates, framingSubstrates, backboneCuts = \
+        ki.buildLayout(preset, panel, input, sourceArea)
 
-    tabCuts = ki.buildTabs(preset["tabs"], panel, substrates,
-        framingSubstrates, ki.frameOffset(preset["framing"]))
-    backboneCuts = ki.buildBackBone(preset["layout"], panel, substrates,
-        ki.frameOffset(preset["framing"]))
-    frameCuts = ki.buildFraming(preset["framing"], panel)
+    useHookPlugins(lambda x: x.afterLayout(panel, substrates))
 
-    ki.buildTooling(preset["tooling"], panel)
-    ki.buildFiducials(preset["fiducials"], panel)
-    ki.buildText(preset["text"], panel)
+    tabCuts = ki.buildTabs(preset, panel, substrates, framingSubstrates)
+
+    useHookPlugins(lambda x: x.afterTabs(panel, tabCuts, backboneCuts))
+
+    frameCuts = ki.buildFraming(preset, panel)
+
+    useHookPlugins(lambda x: x.afterFraming(panel, frameCuts))
+
+    ki.buildTooling(preset, panel)
+    ki.buildFiducials(preset, panel)
+    for textSection in ["text", "text2", "text3", "text4"]:
+        ki.buildText(preset[textSection], panel)
     ki.buildPostprocessing(preset["post"], panel)
 
-    ki.makeTabCuts(preset["cuts"], panel, tabCuts)
-    ki.makeOtherCuts(preset["cuts"], panel, chain(backboneCuts, frameCuts))
+    ki.makeTabCuts(preset, panel, tabCuts)
+    ki.makeOtherCuts(preset, panel, chain(backboneCuts, frameCuts))
+
+    useHookPlugins(lambda x: x.afterCuts(panel))
+
+    ki.buildCopperfill(preset["copperfill"], panel)
 
     ki.setStackup(preset["source"], panel)
     ki.positionPanel(preset["page"], panel)
     ki.setPageSize(preset["page"], panel, board)
 
     ki.runUserScript(preset["post"], panel)
+    useHookPlugins(lambda x: x.finish(panel))
 
     ki.buildDebugAnnotation(preset["debug"], panel)
 
-    panel.save()
+    panel.save(reconstructArcs=preset["post"]["reconstructarcs"],
+               refillAllZones=preset["post"]["refillzones"])
 
 
 @click.command()
@@ -292,7 +349,7 @@ def separate(input, output, source, page, debug, keepannotations):
         ki.positionPanel(preset["page"], panel)
         ki.setPageSize(preset["page"], panel, board)
 
-        panel.save()
+        panel.save(reconstructArcs=True)
     except Exception as e:
         import sys
         sys.stderr.write("An error occurred: " + str(e) + "\n")
