@@ -1,3 +1,7 @@
+from dataclasses import dataclass
+import os
+from typing import Any, List
+from kikit import plugin
 from kikit.units import readLength, readAngle
 from kikit.defs import Layer, EDA_TEXT_HJUSTIFY_T, EDA_TEXT_VJUSTIFY_T, PAPER_SIZES
 
@@ -7,11 +11,19 @@ class PresetError(RuntimeError):
 ANCHORS = ["tl", "tr", "bl", "br", "mt", "mb", "ml", "mr", "c"]
 PAPERS = ["inherit"] + PAPER_SIZES + ["user"]
 
+@dataclass
+class FootprintId:
+    lib: str
+    footprint: str
+
 class SectionBase:
     def __init__(self, isGuiRelevant, description):
         self.description = description
         self.isGuiRelevant = isGuiRelevant
 
+    def validate(self, x: str) -> Any:
+        raise NotImplementedError("Validate was not overridden for SectionBase")
+
 class SLength(SectionBase):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -40,6 +52,48 @@ class SStr(SectionBase):
     def validate(self, x):
         return str(x)
 
+class SPlugin(SectionBase):
+    seq: int = 0
+
+    def __init__(self, pluginType, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.pluginType = pluginType
+
+    def validate(self, x):
+        if x == "none":
+            return None
+        self.seq += 1
+
+        pieces = str(x).rsplit(".", maxsplit=1)
+        if len(pieces) != 2:
+            raise RuntimeError(f"Invalid plugin specification '{x}'")
+        moduleName, pluginName = pieces[0], pieces[1]
+        plugin = self.loadFromFile(moduleName, pluginName) if moduleName.endswith(".py") \
+                 else self.loadFromModule(moduleName, pluginName)
+        if not issubclass(plugin, self.pluginType):
+            raise RuntimeError(f"Invalid plugin type specified, {self.pluginType.__name__} expected")
+        setattr(plugin, "__kikit_preset_repr", x)
+        return plugin
+
+    def loadFromFile(self, file, name):
+        import importlib.util
+
+        if not os.path.exists(file):
+            raise RuntimeError(f"File {file} doesn't exist")
+        spec = importlib.util.spec_from_file_location(
+                f"kikit.user.SPlugin_{self.seq}",
+                file)
+        if spec is None:
+            raise RuntimeError(f"Plugin module '{file}' doesn't exist")
+        pluginModule = importlib.util.module_from_spec(spec)
+        spec.loader.exec_module(pluginModule)
+        return getattr(pluginModule, name)
+
+    def loadFromModule(self, module, name):
+        import importlib
+        pluginModule = importlib.import_module(module)
+        return getattr(pluginModule, name)
+
 class SChoiceBase(SectionBase):
     def __init__(self, vals, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -69,7 +123,7 @@ class SBool(SChoiceBase):
             if sl in ["0", "false", "no"]:
                 return False
             raise PresetError(f"Uknown boolean value '{s}'")
-        raise RuntimeError(f"Got {s}, expected boolean value")
+        raise PresetError(f"Got {s}, expected boolean value")
 
 class SHJustify(SChoiceBase):
     def __init__(self, *args, **kwargs):
@@ -109,10 +163,44 @@ class SLayer(SChoiceBase):
         if isinstance(s, int):
             if s in tuple(item.value for item in Layer):
                 return Layer(s)
-            raise RuntimeError(f"{s} is not a valid layer number")
+            raise PresetError(f"{s} is not a valid layer number")
         if isinstance(s, str):
             return Layer[s.replace(".", "_")]
-        raise RuntimeError(f"Got {s}, expected layer name or number")
+        raise PresetError(f"Got {s}, expected layer name or number")
+
+class SList(SectionBase):
+    def validate(self, x: str) -> Any:
+        return [v.strip() for v in x.split(",")]
+
+class SLayerList(SList):
+    def __init__(self, isGuiRelevant, description, shortcuts={}):
+        super().__init__(isGuiRelevant, description)
+        self._shortcuts = shortcuts
+
+    def validate(self, x: str) -> Any:
+        if x in self._shortcuts:
+            return self._shortcuts[x]
+        return [self.readLayer(x) for x in super().validate(x)]
+
+    def readLayer(self, s: str) -> Layer:
+        if isinstance(s, int):
+            if s in tuple(item.value for item in Layer):
+                return Layer(s)
+            raise PresetError(f"{s} is not a valid layer number")
+        if isinstance(s, str):
+            return Layer[s.replace(".", "_")]
+        raise PresetError(f"Got {s}, expected layer name or number")
+
+class SFootprintList(SList):
+    def validate(self, x: str) -> Any:
+        result: List[FootprintId] = []
+        for v in super().validate(x):
+            s = v.split(":", 1)
+            if len(s) != 2:
+                PresetError(f"'{v}' is not a valid footprint name in the form '<lib>:<footprint>'")
+            result.append(FootprintId(s[0], s[1]))
+        return result
+
 
 def validateSection(name, sectionDefinition, section):
     try:
@@ -135,7 +223,7 @@ def never():
 
 LAYOUT_SECTION = {
     "type": SChoice(
-        ["grid"],
+        ["grid", "plugin"],
         always(),
         "Layout type"),
     "alternation": SChoice(
@@ -152,34 +240,49 @@ LAYOUT_SECTION = {
         never(),
         "Specify the gap between the boards in both direction"),
     "hbackbone": SLength(
-        always(),
-        "The width of vertical and horizontal backbone (0 means no backbone)"),
+        typeIn(["grid"]),
+        "The width of horizontal backbone (0 means no backbone)"),
     "vbackbone": SLength(
-        always(),
-        "The width of vertical and horizontal backbone (0 means no backbone)"),
+        typeIn(["grid"]),
+        "The width of vertical backbone (0 means no backbone)"),
+    "hboneskip": SNum(
+        typeIn(["grid"]),
+        "Skip every given number of horizontal backbones"),
+    "vboneskip": SNum(
+        typeIn(["grid"]),
+        "Skip every given number of vertical backbones"),
     "rotation": SAngle(
         always(),
         "Rotate the boards before placing them in the panel"),
     "rows": SNum(
-        always(),
+        typeIn(["grid"]),
         "Specify the number of boards in the grid pattern"),
     "cols": SNum(
-        always(),
+        typeIn(["grid"]),
         "Specify the number of boards in the grid pattern"),
     "vbonecut": SBool(
-        always(),
+        typeIn(["grid"]),
         "Cut backone in vertical direction"),
     "hbonecut": SBool(
-        always(),
+        typeIn(["grid"]),
         "Cut backone in horizontal direction"),
     "renamenet": SStr(
         always(),
-        "Net renaming pattern"
-    ),
+        "Net renaming pattern"),
     "renameref": SStr(
         always(),
-        "Reference renaming pattern"
-    )
+        "Reference renaming pattern"),
+    "baketext": SBool(
+        always(),
+        "Substitute variables in text elements"
+    ),
+    "code": SPlugin(
+        plugin.LayoutPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppLayout(section):
@@ -222,7 +325,7 @@ def ppSource(section):
 
 TABS_SECTION = {
     "type": SChoice(
-        ["none", "fixed", "spacing", "full", "corner", "annotation"],
+        ["none", "fixed", "spacing", "full", "corner", "annotation", "plugin"],
         always(),
         "Tab type"),
     "vwidth": SLength(
@@ -245,7 +348,20 @@ TABS_SECTION = {
         "Number of tabs in a given direction."),
     "hcount": SNum(
         typeIn(["fixed"]),
-        "Number of tabs in a given direction.")
+        "Number of tabs in a given direction."),
+    "cutout": SLength(
+        typeIn(["fixed"]),
+        "Depth of cutouts into the frame"),
+    "tabfootprints": SFootprintList(
+        typeIn(["annotation"]),
+        "Specify custom footprints that will be used for tab annotations."),
+    "code": SPlugin(
+        plugin.TabsPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppTabs(section):
@@ -255,7 +371,7 @@ def ppTabs(section):
 
 CUTS_SECTION = {
     "type": SChoice(
-        ["none", "mousebites", "vcuts"],
+        ["none", "mousebites", "vcuts", "layer", "plugin"],
         always(),
         "Cut type"),
     "drill": SLength(
@@ -268,7 +384,7 @@ CUTS_SECTION = {
         typeIn(["mousebites", "vcuts"]),
         "Offset cuts into the board"),
     "prolong": SLength(
-        typeIn(["mousebites"]),
+        typeIn(["mousebites", "layer"]),
         "Tangentiall prolong cuts (to cut mill fillets)"),
     "clearance": SLength(
         typeIn(["vcuts"]),
@@ -277,8 +393,15 @@ CUTS_SECTION = {
         typeIn(["vcuts"]),
         "Approximate curves with straight cut"),
     "layer": SLayer(
-        typeIn(["vcuts"]),
-        "Draw V-cuts into a specified layer")
+        typeIn(["vcuts", "layer"]),
+        "Specify layer for the drawings"),
+    "code": SPlugin(
+        plugin.CutsPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppCuts(section):
@@ -286,14 +409,14 @@ def ppCuts(section):
 
 FRAMING_SECTION = {
     "type": SChoice(
-        ["none", "railstb", "railslr", "frame", "tightframe"],
+        ["none", "railstb", "railslr", "frame", "tightframe", "plugin"],
         always(),
         "Framing type"),
     "hspace": SLength(
-        typeIn(["frame", "railslr"]),
+        typeIn(["frame", "railslr", "tightframe"]),
         "Horizontal space between PCBs and the frame"),
     "vspace": SLength(
-        typeIn(["frame", "railstb"]),
+        typeIn(["frame", "railstb", "tightframe"]),
         "Vertical space between PCBs and the frame"),
     "space": SLength(
         never(),
@@ -301,6 +424,14 @@ FRAMING_SECTION = {
     "width": SLength(
         typeIn(["frame", "railstb", "railslr", "tightframe"]),
         "Width of the framing"),
+    "mintotalheight": SLength(
+        typeIn(["frame", "railstb", "tightframe"]),
+        "Minimal height of the panel"
+    ),
+    "mintotalwidth": SLength(
+        typeIn(["frame", "raillr", "tightframe"]),
+        "Minimal width of the panel"
+    ),
     "slotwidth": SLength(
         typeIn(["tightframe"]),
         "Width of the milled slot"),
@@ -314,6 +445,13 @@ FRAMING_SECTION = {
     "fillet": SLength(
         typeIn(["tightframe", "frame", "railslr", "railstb"]),
         "Add fillet to the 4 corners of the panel. Specify fillet radius."),
+    "code": SPlugin(
+        plugin.FramingPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppFraming(section):
@@ -324,7 +462,7 @@ def ppFraming(section):
 
 TOOLING_SECTION = {
     "type": SChoice(
-        ["none", "3hole", "4hole"],
+        ["none", "3hole", "4hole", "plugin"],
         always(),
         "Tooling type"),
     "hoffset": SLength(
@@ -338,7 +476,14 @@ TOOLING_SECTION = {
         "Hole diameter"),
     "paste": SBool(
         typeIn(["3hole", "4hole"]),
-        "Include holes on the paste layer")
+        "Include holes on the paste layer"),
+    "code": SPlugin(
+        plugin.ToolingPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppTooling(section):
@@ -346,7 +491,7 @@ def ppTooling(section):
 
 FIDUCIALS_SECTION = {
     "type": SChoice(
-        ["none", "3fid", "4fid"],
+        ["none", "3fid", "4fid", "plugin"],
         always(),
         "Fiducial type"),
     "hoffset": SLength(
@@ -360,7 +505,14 @@ FIDUCIALS_SECTION = {
         "Diameter of the copper part"),
     "opening": SLength(
         typeIn(["3fid", "4fid"]),
-        "Diameter of the opening")
+        "Diameter of the opening"),
+    "code": SPlugin(
+        plugin.FiducialsPlugin,
+        typeIn(["plugin"]),
+        "Plugin specification as moduleName.pluginName"),
+    "arg": SStr(
+        typeIn(["plugin"]),
+        "String argument for the layout plugin")
 }
 
 def ppFiducials(section):
@@ -404,12 +556,45 @@ TEXT_SECTION = {
     "anchor": SChoice(
         ANCHORS,
         typeIn(["simple"]),
-        "Anchor for positioning the text")
+        "Anchor for positioning the text"),
+    "plugin": SPlugin(
+        plugin.TextVariablePlugin,
+        typeIn(["simple"]),
+        "Plugin for extra text variables")
 }
 
 def ppText(section):
     section = validateSection("text", TEXT_SECTION, section)
 
+COPPERFILL_SECTION = {
+    "type": SChoice(
+        ["none", "solid", "hatched"],
+        always(),
+        "Fill non board areas with copper"),
+    "clearance": SLength(
+        typeIn(["solid", "hatched"]),
+        "Clearance between the fill and boards"),
+    "layers": SLayerList(
+        typeIn(["solid", "hatched"]),
+        "Specify which layer to fill with copper",
+        {
+            "all": Layer.allCu()
+        }),
+    "width": SLength(
+        typeIn(["hatched"]),
+        "Width of hatch strokes"),
+    "spacing": SLength(
+        typeIn(["hatched"]),
+        "Spacing of hatch strokes"),
+    "orientation": SAngle(
+        typeIn(["hatched"]),
+        "Orientation of the strokes"
+    )
+}
+
+def ppCopper(section):
+    section = validateSection("copperfill", COPPERFILL_SECTION, section)
+
 POST_SECTION = {
     "type": SChoice(
         ["auto"],
@@ -417,10 +602,16 @@ POST_SECTION = {
         "Postprocessing type"),
     "copperfill": SBool(
         always(),
-        "Fill unused areas of the panel with copper"),
+        "DEPRECATED, use section copperfill instead. Fill unused areas of the panel with copper"),
     "millradius": SLength(
         always(),
         "Simulate milling operation"),
+    "reconstructarcs": SBool(
+        always(),
+        "Try to reconstruct arcs"),
+    "refillzones": SBool(
+        always(),
+        "Refill all zones in the panel"),
     "script": SStr(
         always(),
         "Specify path to a custom postprocessing script"),
@@ -479,6 +670,10 @@ DEBUG_SECTION = {
     "trace": SBool(
         always(),
         "Print stacktrace"),
+    "drawtabfail": SBool(
+        always(),
+        "Visualize tab building failures"
+    ),
     "deterministic": SBool(
         always(),
         "Make KiCAD IDs deterministic")
@@ -496,6 +691,10 @@ availableSections = {
     "Tooling": TOOLING_SECTION,
     "Fiducials": FIDUCIALS_SECTION,
     "Text": TEXT_SECTION,
+    "Text2": TEXT_SECTION,
+    "Text3": TEXT_SECTION,
+    "Text4": TEXT_SECTION,
+    "Copperfill": COPPERFILL_SECTION,
     "Page": PAGE_SECTION,
     "Post": POST_SECTION,
     "Debug": DEBUG_SECTION,
