@@ -16,6 +16,8 @@ from kikit.common import *
 
 from kikit.defs import STROKE_T, Layer
 
+TABFAIL_VISUAL = False
+
 class PositionError(RuntimeError):
     def __init__(self, message, point):
         super().__init__(message.format(toMm(point[0]), toMm(point[1])))
@@ -40,11 +42,11 @@ def getStartPoint(geom):
             point = geom.GetStart()
         else:
             point = geom.GetStart()
-        return roundPoint(point)
+        return point
 
     if geom.GetShape() in [STROKE_T.S_ARC, STROKE_T.S_CIRCLE]:
-        return roundPoint(geom.GetArcStart())
-    return roundPoint(geom.GetStart())
+        return geom.GetArcStart()
+    return geom.GetStart()
 
 def getEndPoint(geom):
     if isV6():
@@ -56,13 +58,13 @@ def getEndPoint(geom):
             point = geom.GetStart()
         else:
             point = geom.GetEnd()
-        return roundPoint(point)
+        return point
 
     if geom.GetShape() == STROKE_T.S_ARC:
-        return roundPoint(geom.GetArcEnd())
+        return geom.GetArcEnd()
     if geom.GetShape() == STROKE_T.S_CIRCLE:
-        return roundPoint(geom.GetArcStart())
-    return roundPoint(geom.GetEnd())
+        return geom.GetArcStart()
+    return geom.GetEnd()
 
 class CoincidenceList(list):
     def getNeighbor(self, myIdx):
@@ -79,16 +81,16 @@ def findRing(startIdx, geometryList, coincidencePoints, unused):
     """
     unused[startIdx] = False
     ring = [startIdx]
-    if getStartPoint(geometryList[startIdx]) == getEndPoint(geometryList[startIdx]):
+    if roundPoint(getStartPoint(geometryList[startIdx])) == roundPoint(getEndPoint(geometryList[startIdx])):
         return ring
-    currentPoint = getEndPoint(geometryList[startIdx])
+    currentPoint = roundPoint(getEndPoint(geometryList[startIdx]))
     while True:
         nextIdx = coincidencePoints[currentPoint].getNeighbor(ring[-1])
         assert(unused[nextIdx] or nextIdx == startIdx)
-        if currentPoint == getStartPoint(geometryList[nextIdx]):
-            currentPoint = getEndPoint(geometryList[nextIdx])
+        if currentPoint == roundPoint(getStartPoint(geometryList[nextIdx])):
+            currentPoint = roundPoint(getEndPoint(geometryList[nextIdx]))
         else:
-            currentPoint = getStartPoint(geometryList[nextIdx])
+            currentPoint = roundPoint(getStartPoint(geometryList[nextIdx]))
         unused[nextIdx] = False
         if nextIdx == startIdx:
             return ring
@@ -144,7 +146,7 @@ def commonEndPoint(a, b):
     """
     aStart, aEnd = getStartPoint(a), getEndPoint(a)
     bStart, bEnd = getStartPoint(b), getEndPoint(b)
-    if aStart == bStart or aStart == bEnd:
+    if roundPoint(aStart) == roundPoint(bStart) or roundPoint(aStart) == roundPoint(bEnd):
         return aStart
     return aEnd
 
@@ -172,7 +174,7 @@ def approximateArc(arc, endWith):
     last = np.array([outline[-1][0], outline[-1][1]])
     if (np.linalg.norm(end - first) < np.linalg.norm(end - last)):
         outline.reverse()
-    return outline
+    return outline[1:-1]
 
 def approximateBezier(bezier, endWith):
     """
@@ -259,10 +261,10 @@ def toShapely(ring, geometryList):
         shape = geometryList[idxA].GetShape()
         if shape in [STROKE_T.S_ARC, STROKE_T.S_CIRCLE]:
             outline += approximateArc(geometryList[idxA],
-                commonEndPoint(geometryList[idxA], geometryList[idxB]))[1:]
+                commonEndPoint(geometryList[idxA], geometryList[idxB]))
         elif shape in [STROKE_T.S_CURVE]:
             outline += approximateBezier(geometryList[idxA],
-                commonEndPoint(geometryList[idxA], geometryList[idxB]))[1:]
+                commonEndPoint(geometryList[idxA], geometryList[idxB]))
         elif shape in [STROKE_T.S_RECT]:
             outline += createRectangle(geometryList[idxA])
         elif shape in [STROKE_T.S_POLYGON]:
@@ -349,31 +351,49 @@ def substratesFrom(polygons):
         substrates.append(Polygon(polygon.exterior, holes))
     return substrates
 
+def commonCircleKiCAD(a, b, c):
+    """
+    Get a common circle using KiCAD APIs (a little abusive)
+    """
+    arc = pcbnew.PCB_SHAPE()
+    arc.SetShape(STROKE_T.S_ARC)
+    arc.SetArcGeometry(wxPoint(*a), wxPoint(*b), wxPoint(*c))
+    center = [int(x) for x in arc.GetCenter()]
+    mid = [(a[i] + b[i]) // 2 for i in range(2)]
+    if center == mid:
+        return None
+    return roundPoint(center, -8)
+
+def commonCirclePython(a, b, c):
+    """
+    Get a common circle using pure Python implementation
+    """
+    o_l = [(a[i] + b[i]) / 2 for i in range(2)]
+    o_r = [(b[i] + c[i]) / 2 for i in range(2)]
+    d_l = [(b[i] - a[i]) for i in range(2)]
+    d_r = [(c[i] - b[i]) for i in range(2)]
+    n_l = (d_l[1], -d_l[0])
+    n_r = (d_r[1], -d_r[0])
+
+    t_denom = n_l[0] * n_r[1] - n_l[1] * n_r[0]
+    if t_denom == 0:
+        return None # The normal vectors are parallel
+    t = (n_r[0] * (o_l[1] - o_r[1]) + n_r[1] * (o_r[1] - o_l[1])) / t_denom
+    intersection = [o_l[i] + n_l[i] * t for i in range(2)]
+    return roundPoint(intersection)
+
+
 def commonCircle(a, b, c):
     """
     Given three 2D points return (x, y, r) of the circle they lie on or None if
     they lie in a line
     """
-    # Based on http://web.archive.org/web/20161011113446/http://www.abecedarical.com/zenosamples/zs_circle3pts.html
-    m11 = np.matrix([[a[0], a[1], 1],
-                     [b[0], b[1], 1],
-                     [c[0], c[1], 1]])
-    m11d = np.linalg.det(m11)
-    if m11d == 0:
-        return None
-    m12 = np.matrix([[a[0]*a[0] + a[1]*a[1], a[1], 1],
-                     [b[0]*b[0] + b[1]*b[1], b[1], 1],
-                     [c[0]*c[0] + c[1]*c[1], c[1], 1]])
-    m13 = np.matrix([[a[0]*a[0] + a[1]*a[1], a[0], 1],
-                     [b[0]*b[0] + b[1]*b[1], b[0], 1],
-                     [c[0]*c[0] + c[1]*c[1], c[0], 1]])
-    m14 = np.matrix([[a[0]*a[0] + a[1]*a[1], a[0], a[1]],
-                     [b[0]*b[0] + b[1]*b[1], b[0], b[1]],
-                     [c[0]*c[0] + c[1]*c[1], c[0], c[1]]])
-    x = 0.5 * np.linalg.det(m12) / m11d
-    y = -0.5 * np.linalg.det(m13) / m11d
-    r = np.sqrt(x*x + y*y + np.linalg.det(m14) / m11d)
-    return (x, y, r)
+    # We use the KiCAD's API implementation as it seems faster, but v5 doesn't
+    # offer it
+    if isV6():
+        return commonCircleKiCAD(a, b, c)
+    return commonCirclePython(a, b, c)
+
 
 def liesOnSegment(start, end, point, tolerance=fromMm(0.01)):
     """
@@ -428,10 +448,23 @@ def closestIntersectionPoint(origin, direction, outline, maxDistance):
     testLine = LineString([origin, origin + direction * maxDistance])
     inter = testLine.intersection(outline)
     if inter.is_empty:
+        if TABFAIL_VISUAL:
+            import matplotlib.pyplot as plt
+
+            plt.axis('equal')
+            x, y = outline.coords.xy
+            plt.plot(list(map(toMm, x)), list(map(toMm, y)))
+            x, y = testLine.coords.xy
+            plt.plot(list(map(toMm, x)), list(map(toMm, y)))
+            plt.show()
         raise NoIntersectionError(f"No intersection found within given distance", origin)
     origin = Point(origin[0], origin[1])
     if isinstance(inter, Point):
         geoms = [inter]
+    elif isinstance(inter, LineString):
+        # When a linestring is an intersection, we know that the starting or
+        # ending points are the nearest one
+        geoms = [Point(inter.coords[0]), Point(inter.coords[-1])]
     else:
         geoms = inter.geoms
     return min([(g, origin.distance(g)) for g in geoms], key=lambda t: t[1])[0]
@@ -454,7 +487,6 @@ class Substrate:
     def __init__(self, geometryList, bufferDistance=0, revertTransformation=None):
         polygons = [toShapely(ring, geometryList) for ring in extractRings(geometryList)]
         self.substrates = unary_union(substratesFrom(polygons))
-        self.substrates = self.substrates.buffer(bufferDistance)
         if not self.substrates.is_empty:
             self.substrates = shapely.ops.orient(self.substrates)
         self.partitionLine = shapely.geometry.GeometryCollection()
@@ -476,6 +508,7 @@ class Substrate:
         """
         if self.oriented:
             return
+        self.substrates = self.substrates.simplify(SHP_EPSILON)
         self.substrates = shapely.ops.orient(self.substrates)
         self.oriented = True
 
@@ -485,6 +518,13 @@ class Substrate:
         """
         return self.substrates.bounds
 
+    def midpoint(self) -> Tuple[int, int]:
+        """
+        Return a mid point of the bounding box
+        """
+        minx, miny, maxx, maxy = self.substrates.bounds
+        return ((minx + maxx) // 2, (miny + maxy) // 2)
+
     def union(self, other):
         """
         Appends a substrate, polygon or list of polygons. If there is a common
@@ -505,7 +545,7 @@ class Substrate:
         """
         self.substrates = self.substrates.difference(piece)
 
-    def serialize(self):
+    def serialize(self, reconstructArcs=False):
         """
         Produces a list of PCB_SHAPE on the Edge.Cuts layer
         """
@@ -517,26 +557,59 @@ class Substrate:
             raise RuntimeError("Uknown type '{}' of substrate geometry".format(type(self.substrates)))
         items = []
         for polygon in geoms:
-            items += self._serializeRing(polygon.exterior)
+            items += self._serializeRing(polygon.exterior, reconstructArcs)
             for interior in polygon.interiors:
-                items += self._serializeRing(interior)
+                items += self._serializeRing(interior, reconstructArcs)
         return items
 
-    def _serializeRing(self, ring):
-        coords = list(ring.simplify(pcbnew.FromMM(0.001)).coords)
+    def _serializeRing(self, ring, reconstructArcs):
+        coords = ring.coords
         segments = []
-        # ToDo: Reconstruct arcs
         if coords[0] != coords[-1]:
             raise RuntimeError("Ring is incomplete")
-        for a, b in zip(coords, coords[1:]):
-            segment = pcbnew.PCB_SHAPE()
-            segment.SetShape(STROKE_T.S_SEGMENT)
-            segment.SetLayer(Layer.Edge_Cuts)
-            segment.SetStart(wxPoint(*a))
-            segment.SetEnd(wxPoint(*b))
-            segments.append(segment)
+        i = 0
+        while i < len(coords):
+            j = i # in the case the following cycle never happens
+            if reconstructArcs:
+                for j in range(i, len(coords) - 3): # Just walk until there is an arc
+                    cc1 = commonCircle(coords[j], coords[j + 1], coords[j + 2])
+                    cc2 = commonCircle(coords[j + 1], coords[j + 2], coords[j + 3])
+                    if cc1 is None or cc2 is None or cc1 != cc2:
+                        break
+            if j - i > 10:
+                j += 2
+                # Yield a circle
+                a = coords[i]
+                b = coords[(i + j) // 2]
+                c = coords[j]
+                segments.append(self._constructArc(a, b, c))
+                i = j
+            else:
+                # Yield a line
+                a = coords[i]
+                b = coords[(i + 1) % len(coords)]
+                segments.append(self._constructEdgeSegment(a, b))
+                i += 1
         return segments
 
+    def _constructEdgeSegment(self, a, b):
+        segment = pcbnew.PCB_SHAPE()
+        segment.SetShape(STROKE_T.S_SEGMENT)
+        segment.SetLayer(Layer.Edge_Cuts)
+        segment.SetStart(wxPoint(*a))
+        segment.SetEnd(wxPoint(*b))
+        return segment
+
+    def _constructArc(self, a, b, c):
+        """
+        Construct arc based on three points
+        """
+        arc = pcbnew.PCB_SHAPE()
+        arc.SetShape(STROKE_T.S_ARC)
+        arc.SetLayer(Layer.Edge_Cuts)
+        arc.SetArcGeometry(wxPoint(*a), wxPoint(*b), wxPoint(*c))
+        return arc
+
     def boundingBox(self):
         """
         Return bounding box as wxRect
@@ -557,6 +630,9 @@ class Substrate:
         polygons = [Polygon(p.exterior) for p in geoms]
         return unary_union(polygons)
 
+    def exteriorRing(self):
+        return self.substrates.exterior
+
     def boundary(self):
         """
         Return shapely geometry representing the outer ring
@@ -583,7 +659,7 @@ class Substrate:
 
         origin = np.array(origin)
         try:
-            direction = normalize(direction)
+            direction = np.around(normalize(direction))
             sideOriginA = origin + makePerpendicular(direction) * width / 2
             sideOriginB = origin - makePerpendicular(direction) * width / 2
             boundary = self.substrates.exterior
@@ -596,7 +672,7 @@ class Substrate:
                 # There is nothing else to do, return the tab
                 tab = Polygon(list(tabFace.coords) + [sideOriginA, sideOriginB])
                 return tab, tabFace
-            # Span the tab towwards the partition line
+            # Span the tab towards the partition line
             # There might be multiple geometries in the partition line, so try them
             # individually.
             direction = -direction
@@ -605,7 +681,7 @@ class Substrate:
                     partitionSplitPointA = closestIntersectionPoint(splitPointA.coords[0],
                             direction, p, maxHeight)
                     partitionSplitPointB = closestIntersectionPoint(splitPointB.coords[0],
-                        direction, p, maxHeight)
+                            direction, p, maxHeight)
                 except NoIntersectionError: # We cannot span towards the partition line
                     continue
                 if isLinestringCyclic(p):
@@ -620,7 +696,12 @@ class Substrate:
                     partitionFaceCoord = list(partitionFace.coords)
                     if i == 1:
                         partitionFaceCoord = partitionFaceCoord[::-1]
-                    tab = Polygon(list(tabFace.coords) + partitionFaceCoord)
+                    # We offset the tab face a little so we can be sure that we
+                    # penetrate the board substrate. Otherwise, there is a
+                    # numerical instability on small slopes that yields
+                    # artifacts on substrate union
+                    offsetTabFace = [(p[0] - SHP_EPSILON * direction[0], p[1] - SHP_EPSILON * direction[1]) for p in tabFace.coords]
+                    tab = Polygon(offsetTabFace + partitionFaceCoord)
                     return tab, tabFace
             return None, None
         except NoIntersectionError as e:
